import json
from contextlib import asynccontextmanager
from typing import AsyncIterator
from typing import Dict
from typing import Iterator
from typing import List
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

from aiopg.sa import create_engine
from aiopg.sa import SAConnection
from pydantic import BaseModel


def model_to_primitive(
    model: BaseModel,
    without_id: bool = False,
    exclude: Sequence[str] = None,
    keep_python_primitives: bool = False,
) -> Dict:
    """
    Convert pydantic-{model} to dict transforming complex types to primitives (e.g. datetime to str)
    :param model: Pydantic model
    :param without_id: Remove id key from result dict
    :param exclude: List of field to exclude from result dict
    :param keep_python_primitives: If True result dict will have python-primitives (e.g. datetime, Decimal)
    :return: Dict with fields from given model
    """
    exclude_set: Set[Union[int, str]] = set(exclude or [])
    if without_id:
        exclude_set.add("id")

    data: Dict
    if keep_python_primitives:
        data = model.dict(exclude=exclude_set)
    else:
        data = json.loads(model.json(exclude=exclude_set))

    return data


@asynccontextmanager
async def create_async_db_connection(db_url: str) -> SAConnection:
    """Create async db connection via aiopg"""
    async with create_engine(db_url) as engine:
        async with engine.acquire() as connection:
            yield connection


def is_field_equal_to_default(entity: BaseModel, field_name: str) -> bool:
    return getattr(entity, field_name) == entity.__fields__[field_name].default


T = TypeVar("T")
F = TypeVar("F")


async def mixed_zip(first: Iterator[T], second: AsyncIterator[F]) -> AsyncIterator[Tuple[T, F]]:
    """
    A zip-like generator in which the first arg is a synchronous iterator, while the second one
    is an async iterator.
    """
    first_iter = iter(first)
    async for s in second:
        try:
            f = next(first_iter)
        except StopIteration:
            break
        yield f, s


async def aiter_to_list(aiter: AsyncIterator[T]) -> List[T]:
    """Converts an async iterator of entities to list"""
    return [entity async for entity in aiter]



from typing import List, Optional

from .base import BaseDistribution, BaseEnvironment, FilesystemWheel, MemoryWheel, Wheel

__all__ = [
    "BaseDistribution",
    "BaseEnvironment",
    "FilesystemWheel",
    "MemoryWheel",
    "Wheel",
    "get_default_environment",
    "get_environment",
    "get_wheel_distribution",
]


def get_default_environment() -> BaseEnvironment:
    """Get the default representation for the current environment.

    This returns an Environment instance from the chosen backend. The default
    Environment instance should be built from ``sys.path`` and may use caching
    to share instance state accorss calls.
    """
    from .pkg_resources import Environment

    return Environment.default()


def get_environment(paths: Optional[List[str]]) -> BaseEnvironment:
    """Get a representation of the environment specified by ``paths``.

    This returns an Environment instance from the chosen backend based on the
    given import paths. The backend must build a fresh instance representing
    the state of installed distributions when this function is called.
    """
    from .pkg_resources import Environment

    return Environment.from_paths(paths)


def get_directory_distribution(directory: str) -> BaseDistribution:
    """Get the distribution metadata representation in the specified directory.

    This returns a Distribution instance from the chosen backend based on
    the given on-disk ``.dist-info`` directory.
    """
    from .pkg_resources import Distribution

    return Distribution.from_directory(directory)


def get_wheel_distribution(wheel: Wheel, canonical_name: str) -> BaseDistribution:
    """Get the representation of the specified wheel's distribution metadata.

    This returns a Distribution instance from the chosen backend based on
    the given wheel's ``.dist-info`` directory.

    :param canonical_name: Normalized project name of the given wheel.
    """
    from .pkg_resources import Distribution

    return Distribution.from_wheel(wheel, canonical_name)



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.track_modifications import before_models_committed
from quart_sqlalchemy.track_modifications import models_committed

pytest.importorskip("blinker")


async def test_track_modifications(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = True
        db = SQLAlchemy(app)

        class Example(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)
            data = sa.Column(sa.String)

        db.create_all()
        before: list[tuple[t.Any, str]] = []
        after: list[tuple[t.Any, str]] = []

        def before_commit(sender: Quart, changes: list[tuple[t.Any, str]]) -> None:
            nonlocal before
            before = changes

        def after_commit(sender: Quart, changes: list[tuple[t.Any, str]]) -> None:
            nonlocal after
            after = changes

        connect_before = before_models_committed.connected_to(before_commit, app)
        connect_after = models_committed.connected_to(after_commit, app)

        with connect_before, connect_after:
            item = Example()

            db.session.add(item)
            assert not before
            assert not after

            db.session.commit()
            assert len(before) == 1
            assert before[0] == (item, "insert")
            assert before == after

            db.session.remove()
            item = db.session.get(Example, 1)  # type: ignore[assignment]
            item.data = "test"  # type: ignore[assignment]
            db.session.commit()
            assert len(before) == 1
            assert before[0] == (item, "update")
            assert before == after

            db.session.remove()
            item = db.session.get(Example, 1)  # type: ignore[assignment]
            db.session.delete(item)
            db.session.commit()
            assert len(before) == 1
            assert before[0] == (item, "delete")
            assert before == after



from __future__ import annotations

import os.path
import unittest.mock

import pytest
import sqlalchemy as sa
from quart import Quart

from quart_sqlalchemy import SQLAlchemy


async def test_default_engine(app: Quart, db: SQLAlchemy) -> None:
    async with app.app_context():
        assert db.engine is db.engines[None]

    with pytest.raises(RuntimeError):
        assert db.engine


async def test_engine_per_bind(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app)
        assert db.engines["a"] is not db.engine


async def test_config_engine_options(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_ENGINE_OPTIONS"] = {"echo": True}
        db = SQLAlchemy(app)
        assert db.engine.echo


async def test_init_engine_options(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_ENGINE_OPTIONS"] = {"echo": False}
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app, engine_options={"echo": True})
        # init is default
        assert db.engines["a"].echo
        # config overrides init
        assert not db.engine.echo


async def test_config_echo(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_ECHO"] = True
        db = SQLAlchemy(app)
        assert db.engine.echo
        assert db.engine.pool.echo


@pytest.mark.parametrize(
    "value",
    [
        "sqlite://",
        sa.engine.URL.create("sqlite"),
        {"url": "sqlite://"},
        {"url": sa.engine.URL.create("sqlite")},
    ],
)
async def test_url_type(app: Quart, value: str | sa.engine.URL) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_BINDS"] = {"a": value}
        db = SQLAlchemy(app)
        assert str(db.engines["a"].url) == "sqlite://"


def test_no_binds_error(app: Quart) -> None:
    del app.config["SQLALCHEMY_DATABASE_URI"]

    with pytest.raises(RuntimeError) as info:
        SQLAlchemy(app)

    e = "Either 'SQLALCHEMY_DATABASE_URI' or 'SQLALCHEMY_BINDS' must be set."
    assert str(info.value) == e


async def test_no_default_url(app: Quart) -> None:
    async with app.app_context():
        del app.config["SQLALCHEMY_DATABASE_URI"]
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app, engine_options={"echo": True})
        assert None not in db.engines
        assert "a" in db.engines


async def test_sqlite_relative_path(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///test.db"
        db = SQLAlchemy(app)
        db.create_all()
        assert not isinstance(db.engine.pool, sa.pool.StaticPool)
        db_path = db.engine.url.database
        assert db_path.startswith(str(app.instance_path))  # type: ignore[union-attr]
        assert os.path.exists(db_path)  # type: ignore[arg-type]


async def test_sqlite_driver_level_uri(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///file:test.db?uri=true"
        db = SQLAlchemy(app)
        db.create_all()
        db_path = db.engine.url.database
        assert db_path is not None
        assert db_path.startswith(f"file:{app.instance_path}")
        assert os.path.exists(db_path[5:])


async def test_sqlite_memory_defaults(app: Quart) -> None:
    with unittest.mock.patch.object(SQLAlchemy, "_make_engine", autospec=True) as make_engine:
        SQLAlchemy(app)
        options = make_engine.call_args[0][2]
        assert options["poolclass"] is sa.pool.StaticPool
        assert options["connect_args"]["check_same_thread"] is False


async def test_mysql_defaults(app: Quart) -> None:
    with unittest.mock.patch.object(SQLAlchemy, "_make_engine", autospec=True) as make_engine:
        app.config["SQLALCHEMY_DATABASE_URI"] = "mysql:///test"
        SQLAlchemy(app)
        options = make_engine.call_args[0][2]
        assert options["pool_recycle"] == 7200
        assert options["url"].query["charset"] == "utf8mb4"



from typing import Any
from typing import Optional

import sqlalchemy.types as types
from sql_repo.ciphers import OID
from sqlalchemy.sql.type_api import _T
from sqlalchemy.sql.type_api import Dialect


class OIDType(types.TypeDecorator):
    impl = types.Integer
    python_type = OID
    cache_ok = True

    def __init__(self, *args, use_ecc=True, **kwargs):
        self.use_ecc = use_ecc
        super().__init__(*args, **kwargs)

    def process_bind_param(self, value: Optional[_T], dialect: Dialect) -> Any:
        """Receive the target type and convert it to the be stored in the database."""
        if value is None:
            return

        return OID(value).decode()

    def process_result_value(self, value: Optional[Any], dialect: Any) -> Optional[_T]:
        """Receive a raw value from the database and convert it to the target type."""
        if value is None:
            return

        return OID(value)

    def copy(self, **kwargs):
        return OIDType(self._value, **kwargs)



"""SQLAlchemy-based implementation of the repository protocol."""
from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

from sqlalchemy import select, text
from sqlalchemy.exc import IntegrityError, SQLAlchemyError

from sql_repo.repo.base import (
    AbstractRepository,
    RepositoryConflictException,
    RepositoryException,
    BeforeAfter,
    CollectionFilter,
    FilterTypes,
    LimitOffset,
)
from sql_repo.repo.orm import Base


if TYPE_CHECKING:
    from collections import abc
    from datetime import datetime

    from sqlalchemy import Select
    from sqlalchemy.engine import Result
    from sqlalchemy.ext.asyncio import AsyncSession


__all__ = [
    "SQLAlchemyRepository",
    "ModelT",
]

T = TypeVar("T")
ModelT = TypeVar("ModelT", bound="Base")
SQLARepoT = TypeVar("SQLARepoT", bound="SQLAlchemyRepository")
BaseT = TypeVar("BaseT", bound="Base")


@contextmanager
def wrap_sqlalchemy_exception() -> Any:
    """Do something within context to raise a `RepositoryException` chained
    from an original `SQLAlchemyError`.

        >>> try:
        ...     with wrap_sqlalchemy_exception():
        ...         raise SQLAlchemyError("Original Exception")
        ... except RepositoryException as exc:
        ...     print(f"caught repository exception from {type(exc.__context__)}")
        ...
        caught repository exception from <class 'sqlalchemy.exc.SQLAlchemyError'>
    """
    try:
        yield
    except IntegrityError as exc:
        raise RepositoryConflictException from exc
    except SQLAlchemyError as exc:
        raise RepositoryException(f"An exception occurred: {exc}") from exc


class SQLAlchemyRepository(AbstractRepository[ModelT], Generic[ModelT]):
    """SQLAlchemy based implementation of the repository interface."""

    def __init__(
        self, *, session: AsyncSession, select_: Select[tuple[ModelT]] | None = None, **kwargs: Any
    ) -> None:
        """
        Args:
            session: Session managing the unit-of-work for the operation.
            select_: To facilitate customization of the underlying select query.
        """
        super().__init__(**kwargs)
        self.session = session
        self._select = select(self.model_type) if select_ is None else select_

    async def add(self, data: ModelT) -> ModelT:
        """Add `data` to the collection.

        Args:
            data: Instance to be added to the collection.

        Returns:
            The added instance.
        """
        with wrap_sqlalchemy_exception():
            instance = await self._attach_to_session(data)
            await self.session.flush()
            await self.session.refresh(instance)
            self.session.expunge(instance)
            return instance

    async def delete(self, id_: Any) -> ModelT:
        """Delete instance identified by `id_`.

        Args:
            id_: Identifier of instance to be deleted.

        Returns:
            The deleted instance.

        Raises:
            RepositoryNotFoundException: If no instance found identified by `id_`.
        """
        with wrap_sqlalchemy_exception():
            instance = await self.get(id_)
            await self.session.delete(instance)
            await self.session.flush()
            self.session.expunge(instance)
            return instance

    async def get(self, id_: Any) -> ModelT:
        """Get instance identified by `id_`.

        Args:
            id_: Identifier of the instance to be retrieved.

        Returns:
            The retrieved instance.

        Raises:
            RepositoryNotFoundException: If no instance found identified by `id_`.
        """
        with wrap_sqlalchemy_exception():
            self._filter_select_by_kwargs(**{self.id_attribute: id_})
            instance = (await self._execute()).scalar_one_or_none()
            instance = self.check_not_found(instance)
            self.session.expunge(instance)
            return instance

    async def list(self, *filters: FilterTypes, **kwargs: Any) -> list[ModelT]:
        """Get a list of instances, optionally filtered.

        Args:
            *filters: Types for specific filtering operations.
            **kwargs: Instance attribute value filters.

        Returns:
            The list of instances, after filtering applied.
        """
        for filter_ in filters:
            match filter_:
                case LimitOffset(limit, offset):
                    self._apply_limit_offset_pagination(limit, offset)
                case BeforeAfter(field_name, before, after):
                    self._filter_on_datetime_field(field_name, before, after)
                case CollectionFilter(field_name, values):
                    self._filter_in_collection(field_name, values)
                case _:
                    raise RepositoryException(f"Unexpected filter: {filter}")
        self._filter_select_by_kwargs(**kwargs)

        with wrap_sqlalchemy_exception():
            result = await self._execute()
            instances = list(result.scalars())
            for instance in instances:
                self.session.expunge(instance)
            return instances

    async def update(self, data: ModelT) -> ModelT:
        """Update instance with the attribute values present on `data`.

        Args:
            data: An instance that should have a value for `self.id_attribute` that exists in the
                collection.

        Returns:
            The updated instance.

        Raises:
            RepositoryNotFoundException: If no instance found with same identifier as `data`.
        """
        with wrap_sqlalchemy_exception():
            id_ = self.get_id_attribute_value(data)
            # this will raise for not found, and will put the item in the session
            await self.get(id_)
            # this will merge the inbound data to the instance we just put in the session
            instance = await self._attach_to_session(data, strategy="merge")
            await self.session.flush()
            await self.session.refresh(instance)
            self.session.expunge(instance)
            return instance

    async def upsert(self, data: ModelT) -> ModelT:
        """Update or create instance.

        Updates instance with the attribute values present on `data`, or creates a new instance if
        one doesn't exist.

        Args:
            data: Instance to update existing, or be created. Identifier used to determine if an
                existing instance exists is the value of an attribute on `data` named as value of
                `self.id_attribute`.

        Returns:
            The updated or created instance.

        Raises:
            RepositoryNotFoundException: If no instance found with same identifier as `data`.
        """
        with wrap_sqlalchemy_exception():
            instance = await self._attach_to_session(data, strategy="merge")
            await self.session.flush()
            await self.session.refresh(instance)
            self.session.expunge(instance)
            return instance

    def filter_collection_by_kwargs(self, **kwargs: Any) -> None:
        """Filter the collection by kwargs.

        Args:
            **kwargs: key/value pairs such that objects remaining in the collection after filtering
                have the property that their attribute named `key` has value equal to `value`.
        """
        with wrap_sqlalchemy_exception():
            self._select.filter_by(**kwargs)

    @classmethod
    async def check_health(cls, session: AsyncSession) -> bool:
        """Perform a health check on the database.

        Args:
            session: through which we runa check statement

        Returns:
            `True` if healthy.
        """
        return (  # type:ignore[no-any-return]  # pragma: no cover
            await session.execute(text("SELECT 1"))
        ).scalar_one() == 1

    # the following is all sqlalchemy implementation detail, and shouldn't be directly accessed

    def _apply_limit_offset_pagination(self, limit: int, offset: int) -> None:
        self._select = self._select.limit(limit).offset(offset)

    async def _attach_to_session(
        self, model: ModelT, strategy: Literal["add", "merge"] = "add"
    ) -> ModelT:
        """Attach detached instance to the session.

        Args:
            model: The instance to be attached to the session.
            strategy: How the instance should be attached.
                - "add": New instance added to session
                - "merge": Instance merged with existing, or new one added.

        Returns:
            Instance attached to the session - if `"merge"` strategy, may not be same instance
            that was provided.
        """
        match strategy:  # noqa: R503
            case "add":
                self.session.add(model)
                return model
            case "merge":
                return await self.session.merge(model)
            case _:
                raise ValueError("Unexpected value for `strategy`, must be `'add'` or `'merge'`")

    async def _execute(self) -> Result[tuple[ModelT, ...]]:
        return await self.session.execute(self._select)

    def _filter_in_collection(self, field_name: str, values: abc.Collection[Any]) -> None:
        if not values:
            return
        self._select = self._select.where(getattr(self.model_type, field_name).in_(values))

    def _filter_on_datetime_field(
        self, field_name: str, before: datetime | None, after: datetime | None
    ) -> None:
        field = getattr(self.model_type, field_name)
        if before is not None:
            self._select = self._select.where(field < before)
        if after is not None:
            self._select = self._select.where(field > before)

    def _filter_select_by_kwargs(self, **kwargs: Any) -> None:
        for key, val in kwargs.items():
            self._select = self._select.where(getattr(self.model_type, key) == val)



from setuptools import setup


# Metadata goes in pyproject.toml. These are here for GitHub's dependency graph.
setup(
    name="quart-sqlalchemy",
    install_requires=["quart~=0.18.3", "SQLAlchemy~=1.4.46", "anyio~=3.3.4"],
)



from werkzeug.security import check_password_hash
from werkzeug.security import generate_password_hash

from quartr import db


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String, unique=True, nullable=False)
    password_hash = db.Column(db.String, nullable=False)

    posts = db.relationship("Post", back_populates="author")

    def set_password(self, value):
        """Store the password as a hash for security."""
        self.password_hash = generate_password_hash(value)

    # allow password = "..." to set a password
    password = property(fset=set_password)

    def check_password(self, value):
        return check_password_hash(self.password_hash, value)



from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from collections.abc import Collection
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Generic
from typing import TypeVar

from sqlalchemy.orm import DeclarativeBase


T = TypeVar("T")
ModelT = TypeVar("ModelT", bound=DeclarativeBase)
RepoT = TypeVar("RepoT", bound="AbstractRepository")


class RepositoryException(Exception):
    """Base repository exception type."""


class RepositoryConflictException(RepositoryException):
    """Exception for data integrity errors."""


class RepositoryNotFoundException(RepositoryException):
    """Identity not present in collection."""


@dataclass
class BeforeAfter:
    """Data required to filter a query on a `datetime` column."""

    field_name: str
    """Name of the model attribute to filter on."""
    before: datetime | None
    """Filter results where field earlier than this [datetime][datetime.datetime]"""
    after: datetime | None
    """Filter results where field later than this [datetime][datetime.datetime]"""


@dataclass
class CollectionFilter(Generic[T]):
    """Data required to construct a `WHERE ...

    IN (...)` clause.
    """

    field_name: str
    """Name of the model attribute to filter on."""
    values: Collection[T]
    """Values for `IN` clause."""


@dataclass
class LimitOffset:
    """Data required to add limit/offset filtering to a query."""

    limit: int
    """Value for `LIMIT` clause of query."""
    offset: int
    """Value for `OFFSET` clause of query."""


FilterTypes = BeforeAfter | CollectionFilter[Any] | LimitOffset


class AbstractRepository(Generic[T], metaclass=ABCMeta):
    """Interface for persistent data interaction."""

    model_type: type[T]
    """Type of object represented by the repository."""
    id_attribute = "id"
    """Name of the primary identifying attribute on `model_type`."""

    def __init__(self, **kwargs: Any) -> None:
        """Repository constructors accept arbitrary kwargs."""
        super().__init__(**kwargs)

    @abstractmethod
    async def add(self, data: T) -> T:
        """Add `data` to the collection.

        Args:
            data: Instance to be added to the collection.

        Returns:
            The added instance.
        """

    @abstractmethod
    async def delete(self, id_: Any) -> T:
        """Delete instance identified by `id_`.

        Args:
            id_: Identifier of instance to be deleted.

        Returns:
            The deleted instance.

        Raises:
            RepositoryNotFoundException: If no instance found identified by `id_`.
        """

    @abstractmethod
    async def get(self, id_: Any) -> T:
        """Get instance identified by `id_`.

        Args:
            id_: Identifier of the instance to be retrieved.

        Returns:
            The retrieved instance.

        Raises:
            RepositoryNotFoundException: If no instance found identified by `id_`.
        """

    @abstractmethod
    async def list(self, *filters: FilterTypes, **kwargs: Any) -> list[T]:
        """Get a list of instances, optionally filtered.

        Args:
            *filters: Types for specific filtering operations.
            **kwargs: Instance attribute value filters.

        Returns:
            The list of instances, after filtering applied.
        """

    @abstractmethod
    async def update(self, data: T) -> T:
        """Update instance with the attribute values present on `data`.

        Args:
            data: An instance that should have a value for `self.id_attribute` that exists in the
                collection.

        Returns:
            The updated instance.

        Raises:
            RepositoryNotFoundException: If no instance found with same identifier as `data`.
        """

    @abstractmethod
    async def upsert(self, data: T) -> T:
        """Update or create instance.

        Updates instance with the attribute values present on `data`, or creates a new instance if
        one doesn't exist.

        Args:
            data: Instance to update existing, or be created. Identifier used to determine if an
                existing instance exists is the value of an attribute on `data` named as value of
                `self.id_attribute`.

        Returns:
            The updated or created instance.

        Raises:
            RepositoryNotFoundException: If no instance found with same identifier as `data`.
        """

    @abstractmethod
    def filter_collection_by_kwargs(self, **kwargs: Any) -> None:
        """Filter the collection by kwargs.

        Has `AND` semantics where multiple kwargs name/value pairs are provided.

        Args:
            **kwargs: key/value pairs such that objects remaining in the collection after filtering
                have the property that their attribute named `key` has value equal to `value`.

        Raises:
            RepositoryException: if a named attribute doesn't exist on `self.model_type`.
        """

    @staticmethod
    def check_not_found(item_or_none: T | None) -> T:
        """Raise `RepositoryNotFoundException` if `item_or_none` is `None`.

        Args:
            item_or_none: Item to be tested for existence.

        Returns:
            The item, if it exists.
        """
        if item_or_none is None:
            raise RepositoryNotFoundException("No item found when one was expected")
        return item_or_none

    @classmethod
    def get_id_attribute_value(cls, item: T) -> Any:
        """Get value of attribute named as `self.id_attribute` on `item`.

        Args:
            item: Anything that should have an attribute named as `self.id_attribute` value.

        Returns:
            The value of attribute on `item` named as `self.id_attribute`.
        """
        return getattr(item, cls.id_attribute)

    @classmethod
    def set_id_attribute_value(cls, id_: Any, item: T) -> Any:
        """Return the `item` after the ID is set to the appropriate attribute.

        Args:
            id_: Value of ID to be set on instance
            item: Anything that should have an attribute named as `self.id_attribute` value.

        Returns:
            Item with `id_` set to `cls.id_attribute`
        """
        setattr(item, cls.id_attribute, id_)
        return item



import os

import click
from quart import Quart
from quart.cli import with_appcontext
from quart_sqlalchemy import SQLAlchemy

__version__ = (1, 0, 0, "dev")

db = SQLAlchemy()


def create_app(test_config=None):
    """Create and configure an instance of the Quart application."""
    app = Quart(__name__, instance_relative_config=True)

    # some deploy systems set the database url in the environ
    db_url = os.environ.get("DATABASE_URL")

    if db_url is None:
        # default to a sqlite database in the instance folder
        db_url = "sqlite:///quartr.sqlite"

    app.config.from_mapping(
        # default secret that should be overridden in environ or config
        SECRET_KEY=os.environ.get("SECRET_KEY", "dev"),
        SQLALCHEMY_DATABASE_URI=db_url,
    )

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile("config.py", silent=True)
    else:
        # load the test config if passed in
        app.config.update(test_config)

    # initialize Quart-SQLAlchemy and the init-db command
    db.init_app(app)
    app.cli.add_command(init_db_command)

    # apply the blueprints to the app
    from quartr import auth, blog

    app.register_blueprint(auth.bp)
    app.register_blueprint(blog.bp)

    # make "index" point at "/", which is handled by "blog.index"
    app.add_url_rule("/", endpoint="index")

    return app


def init_db():
    db.drop_all()
    db.create_all()


@click.command("init-db")
@with_appcontext
def init_db_command():
    """Clear existing data and create new tables."""
    init_db()
    click.echo("Initialized the database.")



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from enum import Enum
from typing import Annotated
from typing import List
from typing import Optional

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.orm import column_property
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import registry
from sqlalchemy.orm import relationship
from sqlalchemy.orm import WriteOnlyMapped


metadata = MetaData()
reg = registry(metadata=metadata)

intpk = Annotated[int, mapped_column(init=False, primary_key=True)]
intfk = Annotated[int, mapped_column(init=False, nullable=False)]
created_at = Annotated[
    datetime,
    mapped_column(nullable=False, server_default=func.now()),
]
updated_at = Annotated[
    datetime,
    mapped_column(
        nullable=False,
        server_default=func.now(),
        server_onupdate=func.now(),
    ),
]


class Status(Enum):
    PENDING = "pending"
    RECEIVED = "received"
    COMPLETED = "completed"


class Base(DeclarativeBase):
    metadata = metadata


@reg.mapped
class User(MappedAsDataclass, Base):
    """User class will be converted to a dataclass"""

    __tablename__ = "user_account"

    id: Mapped[intpk]
    name: Mapped[str]
    firstname: Mapped[str] = mapped_column(String(50))
    lastname: Mapped[str] = mapped_column(String(50))
    fullname: Mapped[str] = column_property(firstname + " " + lastname)
    version: Mapped[int] = mapped_column()
    created_at: Mapped[datetime] = mapped_column()

    addresses: Mapped[List["Address"]] = relationship(back_populates="user")

    __mapper_args__ = {"version_id_col": version}


class Address(MappedAsDataclass, Base):
    __tablename__ = "address"

    id: Mapped[intpk]
    user_id: Mapped[intfk] = mapped_column(ForeignKey("user.id"))
    email_address: Mapped[str] = mapped_column()
    address_statistics: Mapped[Optional[str]] = mapped_column(Text, deferred=True)
    created_at: Mapped[created_at] = mapped_column()
    updated_at: Mapped[updated_at] = mapped_column()

    user: Mapped["User"] = relationship(back_populates="addresses")


# class Parent(MappedAsDataclass, Base):
#     __tablename__ = "parent"

#     id: Mapped[int] = mapped_column(primary_key=True)
#     children: Mapped[List["Child"]] = relationship(default_factory=list, back_populates="parent")


# class Child(MappedAsDataclass, Base):
#     __tablename__ = "child"

#     id: Mapped[int] = mapped_column(primary_key=True)
#     parent_id: Mapped[int] = mapped_column(ForeignKey("parent.id"))
#     parent: Mapped["Parent"] = relationship(default=None)


# class User:
#     __tablename__ = "user"
#     __sa_dataclass_metadata_key__ = "sa"

#     id: int = field(init=False, metadata={"sa": Column(Integer, primary_key=True)})
#     name: str = field(default=None, metadata={"sa": Column(String(50))})
#     fullname: str = field(default=None, metadata={"sa": Column(String(50))})
#     nickname: str = field(default=None, metadata={"sa": Column(String(12))})

#     addresses: List[Address] = field(default_factory=list, metadata={"sa": relationship("Address")})


# @mapper_registry.mapped
# @dataclass
# class Address:
#     __tablename__ = "address"
#     __sa_dataclass_metadata_key__ = "sa"

#     id: int = field(init=False, metadata={"sa": Column(Integer, primary_key=True)})
#     user_id: int = field(init=False, metadata={"sa": Column(ForeignKey("user.id"))})
#     email_address: Optional[str] = field(default=None, metadata={"sa": Column(String(50))})


# engine = create_engine("sqlite:///:memory:")
# mapper_registry.metadata.create_all(bind=engine)

# user = User(name="Michael")
# user.addresses = [
#     Address(email_address="foo@bar.com"),
#     Address(email_address="bar@foo.com"),
# ]

with Session(engine, expire_on_commit=False) as session, session.begin():
    session.add(user)

print(user)
print(astuple(user))
print(json.dumps(asdict(user), indent=2))



from __future__ import annotations

import sqlalchemy as sa

from quart_sqlalchemy import SQLAlchemy


def test_bind_key_default(db: SQLAlchemy) -> None:
    user_table = db.Table("user", sa.Column("id", sa.Integer, primary_key=True))
    assert user_table.metadata is db.metadata


def test_metadata_per_bind(db: SQLAlchemy) -> None:
    user_table = db.Table(
        "user", sa.Column("id", sa.Integer, primary_key=True), bind_key="other"
    )
    assert user_table.metadata is db.metadatas["other"]


def test_multiple_binds_same_table_name(db: SQLAlchemy) -> None:
    user1_table = db.Table("user", sa.Column("id", sa.Integer, primary_key=True))
    user2_table = db.Table(
        "user", sa.Column("id", sa.Integer, primary_key=True), bind_key="other"
    )

    assert user1_table.metadata is db.metadata
    assert user2_table.metadata is db.metadatas["other"]


def test_explicit_metadata(db: SQLAlchemy) -> None:
    other_metadata = sa.MetaData()
    user_table = db.Table(
        "user",
        other_metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        bind_key="other",
    )
    assert user_table.metadata is other_metadata
    assert "other" not in db.metadatas



from pydantic import BaseModel
from pydantic_appconfig import AppConfigHelper


class MyAppConfig(BaseModel):
    """My app config."""

    test_field_string: str
    test_field_int: int

    class Config:
        """The pydantic config, including title for the JSON schema."""

        title = "MyAppConfig"


config = AppConfigHelper(
    appconfig_application="AppConfig-App",
    appconfig_environment="AppConfig-Env",
    appconfig_profile="AppConfig-Profile",
    max_config_age=15,
    fetch_on_init=True,
    config_schema_model=MyAppConfig,
)
